@@ -20,10 +20,8 @@
 #  but WITHOUT ANY WARRANTY; without even the implied warranty of
 #  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 #  GNU General Public License for more details.
-#  You should have received a copy of the GNU General Public License
-#  along with this program; if not, write to the Free Software
-#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
-#  MA  02110-1301, USA.
+# You should have received a copy of the GNU General Public License
+# along with this program. If not, see <https://www.gnu.org/licenses/>.
 """
 Crazyradio CRTP link driver.
 
@@ -34,10 +32,24 @@ import array
 import binascii
 import collections
 import logging
+import queue
 import re
 import struct
-import sys
 import threading
+import time
+from enum import Enum
+from queue import Queue
+from threading import Semaphore
+from threading import Thread
+from typing import Any
+from typing import Dict
+from typing import Iterable
+from typing import List
+from typing import Optional
+from typing import Tuple
+from typing import Union
+from urllib.parse import parse_qs
+from urllib.parse import urlparse
 
 import cflib.drivers.crazyradio as crazyradio
 from .crtpstack import CRTPPacket
@@ -45,11 +57,6 @@ from .exceptions import WrongUriType
 from cflib.crtp.crtpdriver import CRTPDriver
 from cflib.drivers.crazyradio import Crazyradio
 
-if sys.version_info < (3,):
-    import Queue as queue
-else:
-    import queue
-
 
 __author__ = 'Bitcraze AB'
 __all__ = ['RadioDriver']
@@ -63,68 +70,168 @@ DEFAULT_ADDR_A = [0xe7, 0xe7, 0xe7, 0xe7, 0xe7]
 DEFAULT_ADDR = 0xE7E7E7E7E7
 
 
-class _SharedRadio():
-    """ Manages access to one shared radio
-    """
-
-    def __init__(self, devid):
-        self.radio = Crazyradio(devid=devid)
-        self.lock = threading.Lock()
-        self.usage_counter = 0
+class _RadioCommands(Enum):
+    STOP = 0
+    SEND_PACKET = 1
+    SET_ARC = 2
+    SCAN_SELECTED = 3
+    SCAN_CHANNELS = 4
 
 
-class _RadioManager():
-    """ Radio manager helper class
-     Get a Crazyradio with:
-       radio_manager = _RadioManager(devid)
-     Then use your Crazyradio:
-       with radio_manager as cradio:
-           # cradio is the Crazyradio driver object, it is locked
-     Finally close it when finished.
-      cradio.close()
-    """
-    # Configuration lock. Protects opening and closing Crazyradios
-    _config_lock = threading.Lock()
+class _SharedRadioInstance():
+    def __init__(self, instance_id: int,
+                 cmd_queue: 'Queue[Tuple[int, _RadioCommands, Any]]',
+                 rsp_queue: Queue,
+                 version: float):
+        self._instance_id = instance_id
+        self._cmd_queue = cmd_queue
+        self._rsp_queue = rsp_queue
 
-    _radios = []  # Hardware Crazyradio objects
+        self._channel = 2
+        self._address = [0xe7]*5
+        self._datarate = crazyradio.Crazyradio.DR_2MPS
 
-    def __init__(self, devid, channel=0, datarate=0, address=DEFAULT_ADDR_A):
-        self._devid = devid
+        self._opened = True
+
+        self.version = version
+
+    def set_channel(self, channel: int):
         self._channel = channel
-        self._datarate = datarate
+
+    def set_address(self, address):
         self._address = address
 
-        with _RadioManager._config_lock:
-            if len(_RadioManager._radios) <= self._devid or \
-                    _RadioManager._radios[self._devid] is None:
-                _RadioManager._radios += ((self._devid + 1) -
-                                          len(_RadioManager._radios)) * [None]
-                _RadioManager._radios[self._devid] = _SharedRadio(self._devid)
+    def set_data_rate(self, dr):
+        self._datarate = dr
 
-            _RadioManager._radios[self._devid].usage_counter += 1
+    def send_packet(self, data: List[int]) -> crazyradio._radio_ack:
+        assert(self._opened)
+        self._cmd_queue.put((self._instance_id,
+                             _RadioCommands.SEND_PACKET,
+                             (self._channel,
+                              self._address,
+                              self._datarate,
+                              data)))
+        ack = self._rsp_queue.get()  # type: crazyradio._radio_ack
+        return ack
+
+    def set_arc(self, arc):
+        assert(self._opened)
+        self._cmd_queue.put((self._instance_id,
+                             _RadioCommands.SET_ARC,
+                             arc))
+
+    def scan_selected(self, selected, packet):
+        assert(self._opened)
+        self._cmd_queue.put((self._instance_id,
+                             _RadioCommands.SCAN_SELECTED,
+                             (self._datarate, self._address,
+                              selected, packet)))
+        return self._rsp_queue.get()
+
+    def scan_channels(self, start: int, stop: int, packet: Iterable[int]):
+        assert(self._opened)
+        self._cmd_queue.put((self._instance_id,
+                             _RadioCommands.SCAN_CHANNELS,
+                             (self._datarate, self._address,
+                              start, stop, packet)))
+        return self._rsp_queue.get()
 
     def close(self):
-        with _RadioManager._config_lock:
-            _RadioManager._radios[self._devid].usage_counter -= 1
+        assert(self._opened)
+        self._cmd_queue.put((self._instance_id, _RadioCommands.STOP, None))
+        self._opened = False
 
-            if _RadioManager._radios[self._devid].usage_counter == 0:
-                try:
-                    _RadioManager._radios[self._devid].radio.close()
-                except Exception:
-                    pass
-                _RadioManager._radios[self._devid] = None
 
-    def __enter__(self):
-        _RadioManager._radios[self._devid].lock.acquire()
+class _SharedRadio(Thread):
+    def __init__(self, devid: int):
+        Thread.__init__(self)
+        self._radio = Crazyradio(devid=devid)
+        self._devid = devid
+        self.version = self._radio.version
 
-        _RadioManager._radios[self._devid].radio.set_channel(self._channel)
-        _RadioManager._radios[self._devid].radio.set_data_rate(self._datarate)
-        _RadioManager._radios[self._devid].radio.set_address(self._address)
+        self.name = 'Shared Radio'
 
-        return _RadioManager._radios[self._devid].radio
+        self._cmd_queue = Queue()  # type: Queue[Tuple[int, _RadioCommands, Any]]  # noqa
+        self._rsp_queues = {}  # type: Dict[int, Queue[Any]]
+        self._next_instance_id = 0
 
-    def __exit__(self, type, value, traceback):
-        _RadioManager._radios[self._devid].lock.release()
+        self._lock = Semaphore(1)
+
+        self.setDaemon(True)
+
+        self.start()
+
+    def open_instance(self) -> _SharedRadioInstance:
+        rsp_queue = Queue()
+        with self._lock:
+            instance_id = self._next_instance_id
+            self._rsp_queues[instance_id] = rsp_queue
+            self._next_instance_id += 1
+
+            if self._radio is None:
+                self._radio = Crazyradio(devid=self._devid)
+
+        return _SharedRadioInstance(instance_id,
+                                    self._cmd_queue,
+                                    rsp_queue,
+                                    self.version)
+
+    def run(self):
+        while True:
+            command = self._cmd_queue.get()
+
+            if command[1] == _RadioCommands.STOP:
+                with self._lock:
+                    del self._rsp_queues[command[0]]
+                    if len(self._rsp_queues) == 0:
+                        self._radio.close()
+                        self._radio = None
+            elif command[1] == _RadioCommands.SEND_PACKET:
+                channel, address, datarate, data = command[2]
+                self._radio.set_channel(channel)
+                self._radio.set_address(address)
+                self._radio.set_data_rate(datarate)
+                ack = self._radio.send_packet(data)
+                self._rsp_queues[command[0]].put(ack)
+            elif command[1] == _RadioCommands.SET_ARC:
+                self._radio.set_arc(command[2])
+            elif command[1] == _RadioCommands.SCAN_SELECTED:
+                datarate, address, selected, data = command[2]
+                self._radio.set_data_rate(datarate)
+                self._radio.set_address(address)
+                resp = self._radio.scan_selected(selected, data)
+                self._rsp_queues[command[0]].put(resp)
+            elif command[1] == _RadioCommands.SCAN_CHANNELS:
+                datarate, address, start, stop, packet = command[2]
+                self._radio.set_data_rate(datarate)
+                self._radio.set_address(address)
+                resp = self._radio.scan_channels(start, stop, packet)
+                self._rsp_queues[command[0]].put(resp)
+
+
+class RadioManager:
+    _radios = []  # type: List[Union[_SharedRadio, None]]
+    _lock = Semaphore(1)
+
+    @staticmethod
+    def open(devid: int) -> _SharedRadioInstance:
+        with RadioManager._lock:
+            if len(RadioManager._radios) <= devid:
+                padding = [None] * (devid - len(RadioManager._radios) + 1)
+                RadioManager._radios.extend(padding)
+
+            shared_radio = RadioManager._radios[devid]
+            if not shared_radio:
+                shared_radio = _SharedRadio(devid)
+                RadioManager._radios[devid] = shared_radio
+
+            return shared_radio.open_instance()
+
+    @staticmethod
+    def remove(devid: int):
+        with RadioManager._lock:
+            RadioManager._radios[devid] = None
 
 
 class RadioDriver(CRTPDriver):
@@ -133,7 +240,7 @@ class RadioDriver(CRTPDriver):
     def __init__(self):
         """ Create the link driver """
         CRTPDriver.__init__(self)
-        self._radio_manager = None
+        self._radio = None
         self.uri = ''
         self.link_error_callback = None
         self.link_quality_callback = None
@@ -153,22 +260,21 @@ class RadioDriver(CRTPDriver):
         an error message.
         """
 
-        devid, channel, datarate, address = self.parse_uri(uri)
+        devid, channel, datarate, address, rate_limit = self.parse_uri(uri)
         self.uri = uri
 
-        if self._radio_manager is None:
-            self._radio_manager = _RadioManager(devid,
-                                                channel,
-                                                datarate,
-                                                address)
+        if self._radio is None:
+            self._radio = RadioManager.open(devid)
+            self._radio.set_channel(channel)
+            self._radio.set_data_rate(datarate)
+            self._radio.set_address(address)
         else:
             raise Exception('Link already open!')
 
-        with self._radio_manager as cradio:
-            if cradio.version >= 0.4:
-                cradio.set_arc(_nr_of_arc_retries)
-            else:
-                logger.warning('Radio version <0.4 will be obsoleted soon!')
+        if self._radio.version >= 0.4:
+            self._radio.set_arc(_nr_of_arc_retries)
+        else:
+            logger.warning('Radio version <0.4 will be obsoleted soon!')
 
         # Prepare the inter-thread communication queue
         self.in_queue = queue.Queue()
@@ -176,89 +282,96 @@ class RadioDriver(CRTPDriver):
         self.out_queue = queue.Queue(1)
 
         # Launch the comm thread
-        self._thread = _RadioDriverThread(self._radio_manager,
+        self._thread = _RadioDriverThread(self._radio,
                                           self.in_queue,
                                           self.out_queue,
                                           link_quality_callback,
                                           link_error_callback,
-                                          self)
+                                          self,
+                                          rate_limit)
         self._thread.start()
 
         self.link_error_callback = link_error_callback
 
     @staticmethod
-    def parse_uri(uri):
+    def parse_uri(uri: str):
         # check if the URI is a radio URI
-        if not re.search('^radio://', uri):
+        if not uri.startswith('radio://'):
             raise WrongUriType('Not a radio URI')
 
+        parsed_uri = urlparse(uri)
+        parsed_query = parse_qs(parsed_uri.query)
+        parsed_path = parsed_uri.path.strip('/').split('/')
+
         # Open the USB dongle
-        if not re.search('^radio://([0-9a-fA-F]+)((/([0-9]+))'
-                         '((/(250K|1M|2M))?(/([A-F0-9]+))?)?)?$', uri):
-            raise WrongUriType('Wrong radio URI format!')
-
-        uri_data = re.search('^radio://([0-9a-fA-F]+)((/([0-9]+))'
-                             '((/(250K|1M|2M))?(/([A-F0-9]+))?)?)?$', uri)
-
-        if len(uri_data.group(1)) < 10 and uri_data.group(1).isdigit():
-            devid = int(uri_data.group(1))
+        if len(parsed_uri.netloc) < 10 and parsed_uri.netloc.isdigit():
+            devid = int(parsed_uri.netloc)
         else:
             try:
                 devid = crazyradio.get_serials().index(
-                    uri_data.group(1).upper())
+                    parsed_uri.netloc.upper())
             except ValueError:
                 raise Exception('Cannot find radio with serial {}'.format(
-                    uri_data.group(1)))
+                    parsed_uri.netloc))
 
         channel = 2
-        if uri_data.group(4):
-            channel = int(uri_data.group(4))
+        if len(parsed_path) > 0:
+            channel = int(parsed_path[0])
 
         datarate = Crazyradio.DR_2MPS
-        if uri_data.group(7) == '250K':
-            datarate = Crazyradio.DR_250KPS
-        if uri_data.group(7) == '1M':
-            datarate = Crazyradio.DR_1MPS
-        if uri_data.group(7) == '2M':
-            datarate = Crazyradio.DR_2MPS
+        if len(parsed_path) > 1:
+            if parsed_path[1] == '250K':
+                datarate = Crazyradio.DR_250KPS
+            if parsed_path[1] == '1M':
+                datarate = Crazyradio.DR_1MPS
+            if parsed_path[1] == '2M':
+                datarate = Crazyradio.DR_2MPS
 
         address = DEFAULT_ADDR_A
-        if uri_data.group(9):
-            addr = str(uri_data.group(9))
+        if len(parsed_path) > 2:
+            # We make sure to pad the address with zeros if we do not have the
+            # correct length.
+            addr = '{:0>10}'.format(parsed_path[2])
             new_addr = struct.unpack('<BBBBB', binascii.unhexlify(addr))
             address = new_addr
 
-        return devid, channel, datarate, address
+        rate_limit = None
+        if 'rate_limit' in parsed_query:
+            rate_limit = int(parsed_query['rate_limit'][0])
 
-    def receive_packet(self, time=0):
+        return devid, channel, datarate, address, rate_limit
+
+    def receive_packet(self, wait=0):
         """
         Receive a packet though the link. This call is blocking but will
         timeout and return None if a timeout is supplied.
         """
-        if time == 0:
+        if wait == 0:
             try:
                 return self.in_queue.get(False)
             except queue.Empty:
                 return None
-        elif time < 0:
+        elif wait < 0:
             try:
                 return self.in_queue.get(True)
             except queue.Empty:
                 return None
         else:
             try:
-                return self.in_queue.get(True, time)
+                return self.in_queue.get(True, wait)
             except queue.Empty:
                 return None
 
-    def send_packet(self, pk):
+    def send_packet(self, pk) -> bool:
         """ Send the packet pk though the link """
         try:
             self.out_queue.put(pk, True, 2)
+            return True
         except queue.Full:
             if self.link_error_callback:
                 self.link_error_callback('RadioDriver: Could not send packet'
                                          ' to copter')
+            return False
 
     def pause(self):
         self._thread.stop()
@@ -268,7 +381,7 @@ class RadioDriver(CRTPDriver):
         if self._thread:
             return
 
-        self._thread = _RadioDriverThread(self._radio_manager, self.in_queue,
+        self._thread = _RadioDriverThread(self._radio, self.in_queue,
                                           self.out_queue,
                                           self.link_quality_callback,
                                           self.link_error_callback,
@@ -281,9 +394,9 @@ class RadioDriver(CRTPDriver):
         self._thread.stop()
 
         # Close the USB dongle
-        if self._radio_manager:
-            self._radio_manager.close()
-        self._radio_manager = None
+        if self._radio:
+            self._radio.close()
+        self._radio = None
 
         while not self.out_queue.empty():
             self.out_queue.get()
@@ -292,17 +405,18 @@ class RadioDriver(CRTPDriver):
         self.link_error_callback = None
         self.link_quality_callback = None
 
-    def _scan_radio_channels(self, cradio, start=0, stop=125):
+    def _scan_radio_channels(self, radio: _SharedRadioInstance,
+                             start=0, stop=125):
         """ Scan for Crazyflies between the supplied channels. """
-        return list(cradio.scan_channels(start, stop, (0xff,)))
+        return list(radio.scan_channels(start, stop, (0xff,)))
 
     def scan_selected(self, links):
         to_scan = ()
-        for l in links:
+        for link in links:
             one_to_scan = {}
             uri_data = re.search('^radio://([0-9]+)((/([0-9]+))'
-                                 '(/(250K|1M|2M))?)?$',
-                                 l)
+                                 '(/(250K|1M|2M))?)?',
+                                 link)
 
             one_to_scan['channel'] = int(uri_data.group(4))
 
@@ -318,8 +432,7 @@ class RadioDriver(CRTPDriver):
 
             to_scan += (one_to_scan,)
 
-        with self._radio_manager as cradio:
-            found = cradio.scan_selected(to_scan, (0xFF, 0xFF, 0xFF))
+        found = self._radio.scan_selected(to_scan, (0xFF, 0xFF, 0xFF))
 
         ret = ()
         for f in found:
@@ -338,61 +451,61 @@ class RadioDriver(CRTPDriver):
     def scan_interface(self, address):
         """ Scan interface for Crazyflies """
 
-        if self._radio_manager is None:
+        if self._radio is None:
             try:
-                self._radio_manager = _RadioManager(0)
-            except Exception:
+                self._radio = RadioManager.open(0)
+            except Exception as e:
+                print(e)
                 return []
 
-        with self._radio_manager as cradio:
-            # FIXME: implements serial number in the Crazyradio driver!
-            serial = 'N/A'
+        # FIXME: implements serial number in the Crazyradio driver!
+        serial = 'N/A'
 
-            logger.info('v%s dongle with serial %s found', cradio.version,
-                        serial)
-            found = []
+        logger.info('v%s dongle with serial %s found', self._radio.version,
+                    serial)
+        found = []
 
-            if address is not None:
-                addr = '{:X}'.format(address)
-                new_addr = struct.unpack('<BBBBB', binascii.unhexlify(addr))
-                cradio.set_address(new_addr)
+        if address is not None:
+            # We make sure to pad the address with zeroes to get correct length
+            addr = '{:0>10X}'.format(address)
+            new_addr = struct.unpack('<BBBBB', binascii.unhexlify(addr))
+            self._radio.set_address(new_addr)
 
-            cradio.set_arc(1)
+        self._radio.set_arc(1)
 
-            cradio.set_data_rate(cradio.DR_250KPS)
+        self._radio.set_data_rate(crazyradio.Crazyradio.DR_250KPS)
 
-            if address is None or address == DEFAULT_ADDR:
-                found += [['radio://0/{}/250K'.format(c), '']
-                          for c in self._scan_radio_channels(cradio)]
-                cradio.set_data_rate(cradio.DR_1MPS)
-                found += [['radio://0/{}/1M'.format(c), '']
-                          for c in self._scan_radio_channels(cradio)]
-                cradio.set_data_rate(cradio.DR_2MPS)
-                found += [['radio://0/{}/2M'.format(c), '']
-                          for c in self._scan_radio_channels(cradio)]
-            else:
-                found += [['radio://0/{}/250K/{:X}'.format(c, address), '']
-                          for c in self._scan_radio_channels(cradio)]
-                cradio.set_data_rate(cradio.DR_1MPS)
-                found += [['radio://0/{}/1M/{:X}'.format(c, address), '']
-                          for c in self._scan_radio_channels(cradio)]
-                cradio.set_data_rate(cradio.DR_2MPS)
-                found += [['radio://0/{}/2M/{:X}'.format(c, address), '']
-                          for c in self._scan_radio_channels(cradio)]
+        if address is None or address == DEFAULT_ADDR:
+            found += [['radio://0/{}/250K'.format(c), '']
+                      for c in self._scan_radio_channels(self._radio)]
+            self._radio.set_data_rate(crazyradio.Crazyradio.DR_1MPS)
+            found += [['radio://0/{}/1M'.format(c), '']
+                      for c in self._scan_radio_channels(self._radio)]
+            self._radio.set_data_rate(crazyradio.Crazyradio.DR_2MPS)
+            found += [['radio://0/{}/2M'.format(c), '']
+                      for c in self._scan_radio_channels(self._radio)]
+        else:
+            found += [['radio://0/{}/250K/{:X}'.format(c, address), '']
+                      for c in self._scan_radio_channels(self._radio)]
+            self._radio.set_data_rate(crazyradio.Crazyradio.DR_1MPS)
+            found += [['radio://0/{}/1M/{:X}'.format(c, address), '']
+                      for c in self._scan_radio_channels(self._radio)]
+            self._radio.set_data_rate(crazyradio.Crazyradio.DR_2MPS)
+            found += [['radio://0/{}/2M/{:X}'.format(c, address), '']
+                      for c in self._scan_radio_channels(self._radio)]
 
-        self._radio_manager.close()
-        self._radio_manager = None
+        self._radio.close()
+        self._radio = None
 
         return found
 
     def get_status(self):
         try:
-            radio_manager = _RadioManager(0)
+            radio = RadioManager.open(0)
 
-            with radio_manager as cradio:
-                ver = cradio.version
+            ver = radio.version
 
-            radio_manager.close()
+            radio.close()
 
             return 'Crazyradio version {}'.format(ver)
         except Exception:
@@ -408,11 +521,11 @@ class _RadioDriverThread(threading.Thread):
     Radio link receiver thread used to read data from the
     Crazyradio USB driver. """
 
-    def __init__(self, radio_manager, inQueue, outQueue,
-                 link_quality_callback, link_error_callback, link):
+    def __init__(self, radio, inQueue, outQueue,
+                 link_quality_callback, link_error_callback, link, rate_limit: Optional[int]):
         """ Create the object """
         threading.Thread.__init__(self)
-        self._radio_manager = radio_manager
+        self._radio = radio
         self._in_queue = inQueue
         self._out_queue = outQueue
         self._sp = False
@@ -421,6 +534,7 @@ class _RadioDriverThread(threading.Thread):
         self._retry_before_disconnect = _nr_of_retries
         self._retries = collections.deque()
         self._retry_sum = 0
+        self.rate_limit = rate_limit
 
         self._curr_up = 0
         self._curr_down = 1
@@ -460,36 +574,35 @@ class _RadioDriverThread(threading.Thread):
         dataOut = array.array('B', [0xFF])
         waitTime = 0
         emptyCtr = 0
+        ackStatus = None
 
         # Try up to 10 times to enable the safelink mode
-        with self._radio_manager as cradio:
-            for _ in range(10):
-                resp = cradio.send_packet((0xff, 0x05, 0x01))
-                if resp and resp.data and tuple(resp.data) == (
-                        0xff, 0x05, 0x01):
-                    self._has_safelink = True
-                    self._curr_up = 0
-                    self._curr_down = 0
-                    break
+        for _ in range(10):
+            resp = self._radio.send_packet((0xff, 0x05, 0x01))
+            if resp and resp.data and tuple(resp.data) == (
+                    0xff, 0x05, 0x01):
+                self._has_safelink = True
+                self._curr_up = 0
+                self._curr_down = 0
+                break
         self._link.needs_resending = not self._has_safelink
 
         while (True):
             if (self._sp):
                 break
 
-            with self._radio_manager as cradio:
-                try:
-                    if self._has_safelink:
-                        ackStatus = self._send_packet_safe(cradio, dataOut)
-                    else:
-                        ackStatus = cradio.send_packet(dataOut)
-                except Exception as e:
-                    import traceback
+            try:
+                if self._has_safelink:
+                    ackStatus = self._send_packet_safe(self._radio, dataOut)
+                else:
+                    ackStatus = self._radio.send_packet(dataOut)
+            except Exception as e:
+                import traceback
 
-                    self._link_error_callback(
-                        'Error communicating with crazy radio ,it has '
-                        'probably been unplugged!\nException:%s\n\n%s' % (
-                            e, traceback.format_exc()))
+                self._link_error_callback(
+                    'Error communicating with crazy radio ,it has '
+                    'probably been unplugged!\nException:%s\n\n%s' % (
+                        e, traceback.format_exc()))
 
             # Analyse the in data packet ...
             if ackStatus is None:
@@ -534,6 +647,11 @@ class _RadioDriverThread(threading.Thread):
                 else:
                     waitTime = 0
 
+            # If there is a rate limit setup, sleep here to force the rate limit
+            if self.rate_limit:
+                time.sleep(1.0/self.rate_limit)
+                waitTime = 0
+
             # get the next packet to send of relaxation (wait 10ms)
             outPacket = None
             try:
